
%% initialize codebooks %%

% contains prototypes
w = [];

% contains class labels
c_w = [];

%error set and counter
g=0;
S_p.class = [];
S_p.diff = [];
S_p.vec = [];

% count the number of iterative pruning
prunRun=1;
prunStop = 0;

% build a random selection of N training samples per class
% compute the mean to initialize the prototypes
% N = n_mean_components

mapping = unique(c_trainingset);

for i=1:n_classes

  A = find (c_trainingset == mapping(i));
  
  for j=1:n_proto
    B = [];
    mean_components = randperm (size (A,1));

    if(n_mean_components >0)
      for k=1:n_mean_components
          B (k,:) = trainingset (A (mean_components (k)),:);
      end
    else
        B = trainingset (A,:);
    end

    w (size (w, 1)+1,:) = mean (B);
    c_w (size (c_w, 1)+1,:) = mapping(i);
  end

end


accuracy_training = [];
accuracy_test = [];
costs = [];


%% start training %%

% intialize relevance vector
lambda = ones (1,dim) .* (1/dim);

lambdas = zeros(n_epoch,dim);

for epoch=1:n_epoch

  if verbosity 
    disp(strcat('epoch:',num2str(epoch)))
  end
  
  lambdas(epoch,:)=lambda;
  
  % go through training data in random order
  order = randperm (n_samples);

  % check training speed
  if mod (epoch, 20) == 0
    alpha = alpha_decay * alpha;

    if (epoch > relevance_start)
      epsilon = epsilon_decay * epsilon;
    end
  end

  training_error = 0;
  cost = 0;

  %make it run faster by preallocation
   S_p.class = zeros(1,1000000);
   S_p.diff = zeros(1,1000000);
   S_p.vec = zeros(1000000,size(trainingset,2));
   S_p_ind = 1;
  
  for i=1:n_samples

    % select pattern
    x = trainingset (order (i),:);
    c_x = c_trainingset (order (i));

    % distances to prototypes
    d = sum (((lambda (ones (n_vectors ,1),:))).* ((w - (x (ones (n_vectors, 1),:))).^2), 2);

    % find closest correct und incorrect prototype
    correct = find (c_w == c_x);
    incorrect = find (c_w ~= c_x);

    [dJ, J_index] = min (d (correct));
    [dK, K_index] = min (d (incorrect));

    J = correct (J_index);
    K = incorrect (K_index);

    % update of cost function (linear costfunction)
    mu = (dJ - dK) / (dJ + dK);
    cost = cost + mu;
    dsigma = 1;

    % compute updates for prototypes
    dJdK = (dJ + dK)^2;
    dwJ = dsigma * dK * (x - w (J,:)) / dJdK;  
    dwK = dsigma * dJ * (x - w (K,:)) / dJdK;  

    % perform updates
    w (J,:) = w (J,:) + alpha * dwJ;
    w (K,:) = w (K,:) - alpha * dwK;
    
    % compute relevance update
    if (epoch >= relevance_start)
      if(dJdK==0) %make sure we don't divide by zero
          dlambda = zeros(1,dim);
      else
          dlambda = dJ * ((x - w (J,:)).^2) / dJdK - dK * ((x - w (K,:)).^2) / dJdK;
      end
      if(epoch>=pruning_start) %once we pruned them, don't update them anymore
        dlambda(lambda==0)=0;
      end
        
      lambda = lambda - dsigma * epsilon * dlambda;
      lambda (lambda < 0) = 0;
      lambda = lambda / norm (lambda);
    end
    
    % misclassification
    if dJ > dK
      training_error = training_error + 1;
      g = g+1;

      %S_p.class(length(S_p.class)+1) = c_x;
      %S_p.diff(length(S_p.diff)+1) = dK;
      %S_p.vec(size(S_p.vec,1)+1,:) = x;
      
      %use preallocated space
      S_p.class(S_p_ind) = c_x;
      S_p.diff(S_p_ind) = dK;
      S_p.vec(S_p_ind,:) = x;
      S_p_ind = S_p_ind+1;

    end
  end

  % we sort lambda and delete the least important dimensions
  if (pruning_start > 0 && epoch >= pruning_start && prunStop == 0 && mod(epoch-pruning_start,pruning_freq)==0) %xxx
      
	if(pruning_step*prunRun > dim-pruning_target)   %this checks that we overshoot the target dim, if so reduce prunstep
		target = dim-pruning_target;
		prunStop = 1;
	else
		target = pruning_step*prunRun;
	end

	if verbosity
		disp(strcat('pruning to:',num2str(dim-target)))
	end

	[srt,srt_ind] = sort(lambda);
	lambda(srt_ind(1:target)) = 0;
	prunRun = prunRun+1;

  end

  % add prototypes if needed
  if (g>gmax &&  epoch >= increment_start && mod(epoch-increment_start,increment_freq)==0) %xxx
        
      if verbosity 
          disp('adding prototypes') 
      end
        
      %go through all elements of Sp and add when needed
      cls = unique(S_p.class);
      for k=1:length(cls)
          [value,index] = min(S_p.diff(S_p.class == cls(k)));
          w (size (w, 1)+1,:) = S_p.vec(index,:);
          c_w (size (c_w, 1)+1,:) = S_p.class(index);
      end
      n_vectors = size(w,1);
      %S_p.class = [];
      %S_p.diff = [];
      %S_p.vec = [];
      
      %with preallocated mem
      S_p.class = zeros(1,1000000);
      S_p.diff = zeros(1,1000000);
      S_p.vec = zeros(1000000,80);
      S_p_ind = 1;
      g=0;

  end
  
  % compute testerror
  test_error = 0;
  for i=1:n_testsamples
    
    x = testset (i,:);
    c_x = c_testset (i);
    
    d = sqrt (sum (((lambda (ones (n_vectors ,1),:))).* ((w - x (ones (n_vectors, 1),:)).^2), 2));
    [dJ, J_index] = min (d);
    test_error = test_error + (c_x ~= c_w (J_index));
    
  end


  accuracy_training (epoch) = 1 - training_error/n_samples;
  accuracy_test (epoch) = 1 - test_error/n_testsamples;
  costs (epoch) = cost;

end




subplot (3,1,1)
plot ([accuracy_training' accuracy_test']);
xlabel ('# epochs');
ylabel ('accuracy');
legend ('training', 'test', 0);
subplot (3,1,2)
plot (costs)
xlabel ('# epochs');
ylabel ('costfunction');
subplot (3,1,3);
bar (lambda);
title ('Relevance Factors')
xlabel ('Dimension');
ylabel ('\lambda');
